class EmbeddedDocumentMixin:
    _fields = {}
    _db_name_map = {}
    _dirty_fields = {}

    def __init__(self, **kwargs):
        self._values = {}
        for name, value in kwargs.items():
            if name in self._fields:
                setattr(self, name, value)
        self._make_clean()

    def to_mongo(self):
        data = {}
        for field in self._fields.values():
            value = field.to_mongo(
                getattr(self, field.name, None))
            if value is not None:
                data[field.db_name] = value
        return data

    def to_dict(self):
        dct = {}
        for name, field in self._fields.items():
            dct[name] = getattr(self, name, None)
        return dct

    @classmethod
    async def from_mongo(cls, dct, resolver):

        kwargs = {}
        for db_name, value in dct.items():
            field_name = cls._db_name_map[db_name]
            field = cls._fields[field_name]
            value = await field.from_mongo(value, resolver)
            kwargs[field_name] = value

        return cls(**kwargs)

    @property
    def is_valid(self):
        for field in self._fields.values():
            if not field.validate(getattr(self, field.name, None)):
                return False
        return True

    @property
    def is_dirty(self):
        return len(self._dirty_fields) > 0

    def _make_clean(self):
        self._dirty_fields = set()

    @property
    def _identity(self):
        return getattr(self, self._db_name_map['_id'], None)

    @_identity.setter
    def _identity(self, value):
        return setattr(self, self._db_name_map['_id'], value)

    def __eq__(self, other):
        for name in self._fields.keys():
            if getattr(self, name, None) != getattr(other, name, None):
                return False
        return True

    def __repr__(self):
        name = self.__class__.__name__
        args = ','.join([
            name + '=' + repr(getattr(self, name, None))
            for name, field in self._fields.items()])
        return f"{name}({args})"
        # return str(self.to_dict())


class DocumentMixin(EmbeddedDocumentMixin):

    @classmethod
    def qs(self, db):
        raise Exception('This method is replaced by the metaclass')

    def before_create(self):
        pass

    def before_update(self):
        pass



from .instance_query_set import InstanceQuerySet
from .class_query_set import ClassQuerySet


class QuerySet:

    def __get__(self, instance, owner):
        if instance is None:
            return lambda db: ClassQuerySet(owner, db)
        else:
            return lambda db: InstanceQuerySet(instance, db)



from .instance_query_set import InstanceQuerySet
from ..documents.document_iterator import DocumentIterator


class ClassQuerySet(object):

    def __init__(self, document_class, db):
        self.document_class = document_class
        self.db = db

    @property
    def collection(self):
        return self.db[self.document_class.__collection__]

    async def resolve(self, document_class, value):
        qs = document_class.qs(self.db)
        document = await qs.get(value)
        return document

    async def create(self, **kwargs):
        document = self.document_class(**kwargs)
        query_set = InstanceQuerySet(document, self.db)
        return await query_set.create()

    async def get(self, id):
        kwargs = {self.document_class._db_name_map['_id']: {'$eq': id}}
        value = await self.find_one(**kwargs)
        return value

    async def find_one(self, **kwargs):
        query = self._to_query(**kwargs)
        result = await self.collection.find_one(query)
        if not result:
            return None
        value = await self.document_class.from_mongo(result, self.resolve)
        return value

    def find(self, **kwargs):
        query = self._to_query(**kwargs)
        return DocumentIterator(self.collection.find(query), self.document_class, self.db)

    async def count_documents(self, **kwargs):
        query = self._to_query(**kwargs)
        return await self.collection.count_documents(query)

    def _to_query_op(self, field, operation):
        if isinstance(operation, dict):
            query = {}
            for operator, value in operation.items():
                if operator in (
                        '$exists', '$type', '$search', '$language', '$caseSensitive', '$diacriticSensitive', '$size',
                        '$regex', '$options', '$mod'):
                    query[operator] = value
                elif operator in ('$in', '$nin', '$all'):
                    query[operator] = [field.to_mongo(item) for item in value]
                else:
                    query[operator] = field.to_mongo(value)
            return query
        else:
            return operation

    def _to_query_clause(self, key, value):
        if key in ('$and', '$or', '$nor', '$elemMatch', '$text', '$expr'):
            return key, [self._to_query(**item) for item in value]
        elif key in ('$not'):
            return key, self._to_query(**value)
        else:
            return self.document_class._fields[key].db_name, self._to_query_op(self.document_class._fields[key], value)

    def _to_query(self, **kwargs):
        return dict(self._to_query_clause(k, v) for k, v in kwargs.items())

    async def drop(self):
        await self.collection.drop()

    async def ensure_indices(self):
        keys = [
            (
                self.document_class._fields[field_name].db_name,
                self.document_class._fields[field_name].sort_order
            )
            for field_name in self.document_class._indices
        ]
        if len(keys) > 0:
            await self.collection.create_index(keys, unique=True)

    async def insert_many(self, *documents):
        data = [document.to_mongo() for document in documents]
        ret = await self.collection.insert_many(data)
        for id, document in zip(ret.inserted_ids, documents):
            document._identity = str(id)
        return documents



import types

_CALLABLE_TYPES = (
    types.FunctionType,
    types.BuiltinFunctionType,
    types.MethodType,
    types.BuiltinMethodType,
    types.LambdaType
)


def is_callable(value):
    return isinstance(value, _CALLABLE_TYPES)



from .field import Field
from datetime import datetime


class DateTimeField(Field):

    def validate(self, value):
        if value is not None:
            if not isinstance(value, datetime):
                return False

        return super().validate(value)



from asyncio import Future
from pymongo import ASCENDING


class Field:

    def __init__(self, name=None, db_name=None, default=None, required=False, unique=False, sort_order=ASCENDING, before_set=None, after_get=None):
        self.name = name
        self.db_name = db_name
        self.default = default
        self.required = required
        self.unique = unique
        self.sort_order = sort_order
        self.before_set = before_set
        self.after_get = after_get

    def to_mongo(self, value):
        return value

    async def from_mongo(self, value, resolver=None):
        return value

    def is_empty(self, value):
        return value is None

    def validate(self, value):
        return not (self.required and self.is_empty(value))

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.name not in instance._values:
            if callable(self.default):
                instance._values[self.name] = self.default()
            else:
                instance._values[self.name] = self.default

        value = instance._values[self.name]

        return value if not self.after_get else self.after_get(value)

    def __set__(self, instance, value):
        instance._values[self.name] = value if not self.before_set else self.before_set(
            value)
        instance._dirty_fields.add(self.name)



from setuptools import setup, find_packages
from codecs import open
from os import path

from motorodm import __version__ as version

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Load the requirements from the text file.
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [x.strip() for x in f.readlines() if x.strip()]

setup(
    name='motorodm',
    version=version,
    description=("An ODM for motor."),
    lond_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: AsyncIO',
    ],
    author="Rob Blackbourn",
    author_email="rob.blackbourn@gmail.com",
    url='https://github.com/rob-blackbourn/motorodm/',
    license='Apache 2',
    keywords='motor odm mongo',
    packages=find_packages(exclude=['tests', 'examples']),
    python_requires='~=3.6',
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-asyncio']
)



from .field import Field


class StringField(Field):

    def __init__(self, regex=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.regex = regex

    def validate(self, value):
        if value is not None:
            if not isinstance(value, str):
                return False

            if self.regex and not self.regex.match(value):
                return False

        return super().validate(value)

    def is_empty(self, value):
        return value is None or value == ""



from bson.json_util import loads, dumps
from .field import Field


class JsonField(Field):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_mongo(self, value):
        return dumps(value) if value else None

    async def from_mongo(self, value, resolver=None):
        return loads(value) if value else None



class InstanceQuerySet(object):

    def __init__(self, document, db):
        self.document = document
        self.db = db

    @property
    def collection(self):
        return self.db[self.document.__collection__]

    async def save(self):
        if self.document._identity:
            return await self.update()
        else:
            return await self.create()

    async def create(self):
        self.document.before_create()
        data = self.document.to_mongo()
        ret = await self.collection.insert_one(data)
        self.document._identity = str(ret.inserted_id)

        self.document._make_clean()

        return self.document

    async def delete(self):
        await self.collection.delete_one({'_id': self.document._identity})

    async def update(self):
        self.document.before_update()
        data = self.document.to_mongo()
        id = data.pop('_id')
        updates = {
            db_name: data[db_name]
            for db_name in map(
                lambda field_name: self.document._db_name_map[field_name],
                self.document._dirty_fields
            )
        }
        await self.collection.update_one({'_id': id}, {'$set': updates})

        self.document._make_clean()

        return self.document
